/***************************************************************************
                         log.c  -  description
                             -------------------
    begin             : 23.03.2008
    last modified     : 17.06.2020    
    copyright         : (C) 2008 - 2020 by MHS-Elektronik GmbH & Co. KG, Germany
    author            : Klaus Demlehner, klaus@mhs-elektronik.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software, you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License           *
 *   version 2.1 as published by the Free Software Foundation.             *
 *                                                                         *
 ***************************************************************************/

/**
    Library to talk to Tiny-CAN devices. You find the latest versions at
       http://www.tiny-can.com/
**/
#include "global.h"
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <time.h>
#include "global.h"
#include "can_types.h"
#include "usb_hlp.h"
#include "util.h"
#include "log.h"


/**************************************************************************/
/*                         V A R I A B L E N                              */
/**************************************************************************/
static FILE *LogFile;
static DRV_LOCK_TYPE *LogLock = NULL;
static uint32_t LogCreated;
static uint32_t LogFlags;

uint32_t LogEnable;


/**************************************************************************/
/*                        C O N S T A N T E N                             */
/**************************************************************************/
#define DRV_STATUS_MAX 9

static const char *DrvStatusStrings[] = {
  "NOT_LOAD",       // 0  Die Treiber DLL wurde noch nicht geladen
  "NOT_INIT",       // 1  Treiber noch nicht Initialisiert (Funktion "CanInitDrv" noch nicht aufgerufen)
  "INIT",           // 2  Treiber erfolgrich Initialisiert
  "PORT_NOT_OPEN",  // 3  Die Schnittstelle wurde nicht geffnet
  "PORT_OPEN",      // 4  Die Schnittstelle wurde geffnet
  "DEVICE_FOUND",   // 5  Verbindung zur Hardware wurde Hergestellt
  "CAN_OPEN",       // 6  Device wurde geffnet und erfolgreich Initialisiert
  "CAN_RUN_TX",     // 7  CAN Bus RUN nur Transmitter (wird nicht verwendet !)
  "CAN_RUN",        // 8  CAN Bus RUN
  "UNBEKANNT"};

#define CAN_STATUS_MAX 5

static const char *CanStatusStrings[] = {
  "OK",          // 0  Ok
  "ERROR",       // 1  CAN Error
  "WARNING",     // 2  Error warning
  "PASSIV",      // 3  Error passiv
  "BUS_OFF",     // 4  Bus Off
  "UNBEKANNT"};  // 5  Status Unbekannt

#define FIFO_STATUS_MAX 4

static const char *FifoStatusStrings[] = {
  "OK",                 // 0  Fifo-Status: Ok
  "HW_OVERRUN",         // 1  Fifo-Status: Hardware Fifo berlauf
  "SW_OVERRUN",         // 2  Fifo-Status: Software Fifo berlauf
  "HW_SW_OVERRUN",      // 3  Fifo-Status: Hardware & Software Fifo berlauf
  "UNBEKANNT"};         // 4  Fifo-Status: Unbekannt

static const char *CanMsgBusStatusStrings[] = {
  "Bus Ok",
  "Error Warn.",
  "Error Passiv",
  "Bus Off!"};
                   
static const char *CanMsgBusErrorStrings[] = {
  "-",                                           
  "Stuff Error",
  "Form Error",
  "Ack Error",
  "Bit1 Error",
  "Bit0 Error",
  "CRC Error"};


#define VALUE_TYPE_DESC_SIZE 22

struct TValueTypeDesc
  {
  uint32_t Type; 
  const char *Bezeichner;
  };

static const struct TValueTypeDesc ValueTypeDesc[VALUE_TYPE_DESC_SIZE] = {
  {VT_BYTE  ,   "BYTE"},
  {VT_UBYTE ,   "UBYTE"},
  {VT_WORD  ,   "WORD"},
  {VT_UWORD ,   "UWORD"},
  {VT_LONG  ,   "LONG"},
  {VT_ULONG ,   "ULONG"},
  {VT_HBYTE ,   "HBYTE"},
  {VT_HWORD ,   "HWORD"},
  {VT_HLONG ,   "HLONG"},
  {VT_REVISION, "REVISION"},       
  {VT_DATE,     "DATE"},  
  {VT_STRING,   "STRING"}};
  
  
/**************************************************************************/
/*                        F U N K T I O N E N                             */
/**************************************************************************/

static void GetLogTimeStamp(char *buffer)
{
time_t time_now;
#if _MSC_VER >= 1500
struct tm time_info;
#else
struct tm *time_info;
#endif

time(&time_now);
#if _MSC_VER >= 1500
localtime_s(&time_info, &time_now);
strftime(buffer, 30, "%H:%M:%S %d.%m.%Y", &time_info);
#else
time_info = localtime(&time_now);
strftime(buffer, 30, "%H:%M:%S %d.%m.%Y", time_info);
#endif
}


static void LogTimeStamp(void) 
{
char buffer[30];

if ((LogFlags & LOG_WITH_TIME)) 
  {
  GetLogTimeStamp((char *)&buffer);
  fprintf(LogFile, "%s  ", buffer);
  }
}

/*
******************** LogInit ********************
Funktion  : Initialisiert den Treiber

Eingaben  : keine

Ausgaben  : keine
*/
void LogInit(void)
{
LogLock = NULL;
LogFile = NULL;
LogEnable = 0;
LogCreated = 0;
}


/*
******************** LogCreate ********************
Funktion  :

Eingaben  : keine

Ausgaben  : keine
*/
int32_t LogCreate(void)
{
if (!LogCreated)
  {
  DRV_LOCK_INIT(LogLock);
  LogCreated = 1;
  }
return(0);
}

/*
******************** LogDestroy ********************
Funktion  :

Eingaben  : keine

Ausgaben  : keine
*/
void LogDestroy(void)
{
if (LogCreated)
  {
  LogClose();
  DRV_LOCK_DESTROY(LogLock);
  LogCreated = 0;
  }
}


/*
******************** LogOpen ********************
Funktion  : Log File ffnen

Eingaben  : filename => Log Filename

Ausgaben  : result => 0 = kein Fehler
                      -1 = Fehler beim ffnen/anlegen des Log Files
*/
int32_t LogOpen(char *filename, uint32_t flags)
{
if (!LogCreated)
  return(-1);
LogClose();
DRV_LOCK_ENTER(LogLock);
LogFlags = flags;
//LogFile = fopen(filename, "a"); //O_WRONLY|O_APPEND|O_CREAT);
#if _MSC_VER >= 1500
if (fopen_s(&LogFile, filename, "w+"))
  LogFile = NULL;
#else
LogFile = fopen(filename, "w+");
#endif
//strncpy(LogFileName, filename, MAX_FILENAME_SIZE);
DRV_LOCK_LEAVE(LogLock);
if (LogFile)
  {
  LogEnable = 1;
  return(0);
  }
else
  return(-1);
}


/*
******************** LogClose ********************
Funktion  : Log File schlieen

Eingaben  : keine

Ausgaben  : keine
*/
void LogClose(void)
{
if (!LogCreated)
  return;
DRV_LOCK_ENTER(LogLock);
LogEnable = 0;
if (LogFile)
  {
  fflush(LogFile);
  (void)fclose(LogFile);
  }
LogFile = NULL;
//LogFileName[0] = (char)0;
DRV_LOCK_LEAVE(LogLock);
}


/*
******************** LogPrintf ********************
Funktion  : Einen "printf" formatierten String in das Log File schreiben

Eingaben  : log_type =>
            format_str => printf Format String
            ... => Printf Parameter

Ausgaben  : keine
*/
void LogPrintf(uint32_t log_type, uint32_t log_flags, const char *format_str, ...)
{
va_list ap;

if ((LogEnable) && (log_type & LogFlags))
  {
  if ((!log_flags) || (log_flags & LOG_LOCK))
    DRV_LOCK_ENTER(LogLock);
  LogTimeStamp();  
  va_start(ap, format_str);
  vfprintf(LogFile, format_str, ap);
  va_end(ap);
  fprintf(LogFile, "\n");
#ifndef LOG_FILE_MODE_BUFFERED
  if (!(LogFlags & LOG_DISABLE_SYNC))
    fflush(LogFile);
#endif
  if ((!log_flags) || (log_flags & LOG_UNLOCK))
    DRV_LOCK_LEAVE(LogLock);
  }
}


static void LogWriteMessages(struct TCanMsg *msg, int32_t count, int32_t enable_source)
{
char *ptr;
unsigned char ch, hex, dev;
int32_t cnt, i;
const char *type_str;
char id_str[10];
char dlc_str[3];
char msg_str[30];

for (;count; count--)
  {
  // Message Format
  if ((msg->MsgRTR) && (msg->MsgEFF))
    type_str = "EFF/RTR";
  else if (msg->MsgEFF)
    type_str = "EFF    ";
  else if (msg->MsgRTR)
    type_str = "STD/RTR";
  else
    type_str = "STD    ";
  // ID
  if (msg->MsgEFF)
   safe_sprintf(id_str, 10, "%08X", msg->Id);
  else
   safe_sprintf(id_str, 10, "     %03X", msg->Id);
  // Dlc
  cnt = msg->MsgLen;
  if (cnt > 8)
    cnt = 8;
  safe_sprintf(dlc_str, 2, "%1u", (uint32_t)cnt);
  // Daten
  if (msg->MsgRTR)
    cnt = 0;
  ptr = msg_str;
  i=0;
  for (; cnt; cnt--)
    {
    ch = msg->MsgData[i++];
    hex = ch >> 4;
    if (hex > 9)
      *ptr++ = 55 + hex;
    else
      *ptr++ = '0' + hex;
    hex = ch & 0x0F;
    if (hex > 9)
      *ptr++ = 55 + hex;
    else
      *ptr++ = '0' + hex;
    *ptr++ = ' ';
    }
  *ptr = '\0';  
  if (enable_source)
    {
    ch = msg->MsgSource & 0xF;
    dev = (msg->MsgSource >> 4) & 0x0F;
    fprintf(LogFile, "      Dev:%2u Ch:%2u  %8u.%8u  %s | %s | %s | %s\n", 
            dev, ch, msg->Time.Sec, msg->Time.USec, type_str, id_str, dlc_str, msg_str);
    }
  else
    {        
    fprintf(LogFile, "      %8u.%8u  %s | %s | %s | %s\n", 
            msg->Time.Sec, msg->Time.USec, type_str, id_str, dlc_str, msg_str);
    }        
  msg++;
  }
}


static void LogWriteFdMessages(struct TCanFdMsg *msg, int32_t count, uint32_t flags)
{
char *ptr;
unsigned char ch, hex, dev, first, chars;
int32_t cnt, i, sec, usec;
const char *type_str, *fd_str, *space_str;
char id_str[10];
char dlc_str[4];
char msg_str[30];

for (; count; count--)
  {
  if (msg->MsgFD)
    {
    if (msg->MsgBRS)
      fd_str = "FD/BRS";
    else
      fd_str = "FD    "; 
    }
  else
    fd_str = "      ";       
  // Message Format
  if ((msg->MsgRTR) && (msg->MsgEFF))
    type_str = "EFF/RTR";
  else if (msg->MsgEFF)
    type_str = "EFF    ";
  else if (msg->MsgRTR)
    type_str = "STD/RTR";
  else
    type_str = "STD    ";
  // ID
  if (msg->MsgEFF)
   safe_sprintf(id_str, 10, "%08X", msg->Id);
  else
   safe_sprintf(id_str, 10, "     %03X", msg->Id);
  // Dlc
  cnt = msg->MsgLen;
  safe_sprintf(dlc_str, 3, "%2u", (uint32_t)cnt);
  // Daten
  if (msg->MsgRTR)
    cnt = 0;
  if (flags & 0x02)
    {
    sec = 0;
    usec = 0;
    }
  else
    {  
    sec = msg->Time.Sec;
    usec = msg->Time.USec;
    }
  if (DriverInit == TCAN_DRV_FD_INIT)
    {    
    if (flags & 0x01)
      {
      space_str = "                                                                          ";
      ch = msg->MsgSource & 0xF;
      dev = (msg->MsgSource >> 4) & 0x0F;
      fprintf(LogFile, "      Dev:%2u Ch:%2u  %8u.%8u  %s %s | %s | %s | ", 
              dev, ch, sec, usec, fd_str, type_str, id_str, dlc_str);
      }
    else
      {    
      space_str = "                                                          ";
      fprintf(LogFile, "      %8u.%8u  %s %s | %s | %s | ", 
              sec, usec, fd_str, type_str, id_str, dlc_str);
      }
    }
  else
    {
    space_str = " ";
    if (flags & 0x01)
      {
      ch = msg->MsgSource & 0xF;
      dev = (msg->MsgSource >> 4) & 0x0F;
      fprintf(LogFile, "      Dev:%2u Ch:%2u  %8u.%8u  %s | %s | %s | ", 
              dev, ch, sec, usec, type_str, id_str, dlc_str);
      }
    else
      {    
      fprintf(LogFile, "      %8u.%8u  %s | %s | %s | ", 
              sec, usec, type_str, id_str, dlc_str);
      }
    }
  first = 1;  
  i = 0;
  while (cnt)
    {  
    ptr = msg_str;    
    for (chars = 0; cnt; cnt--)
      {
      if (++chars > 8)
        break;
      ch = msg->MsgData[i++];
      //ptr += sprintf(ptr, "%02x ", ch);
      hex = ch >> 4;
      if (hex > 9)
        *ptr++ = 55 + hex;
      else
        *ptr++ = '0' + hex;
      hex = ch & 0x0F;
      if (hex > 9)
        *ptr++ = 55 + hex;
      else
        *ptr++ = '0' + hex;
      *ptr++ = ' ';            
      }      
    *ptr = '\0';
    if (chars)
      {
      if (first)
        {
        first = 0;    
        fprintf(LogFile, "%s\n", msg_str);
        }
      else    
        fprintf(LogFile, "%s%s\n", space_str, msg_str);
      }  
    }
  if (first)
    fprintf(LogFile, "\n");                  
  msg++;
  }
}


static void LogWriteFdErrMsg(struct TCanFdMsg *msg, int32_t enable_source)
{
uint8_t ch, dev, err_nr, bus_stat;
uint32_t lost_msgs;
const char *can_drv_stat;

ch = msg->MsgSource & 0x0F;
dev = (msg->MsgSource >> 4) & 0x0F;
if (msg->MsgErr)
  {  
  err_nr = msg->MsgData[0];
  bus_stat = msg->MsgData[1] & 0x0F;
  if (msg->MsgData[1] & 0x10)
    can_drv_stat = " BUS-FAILURE";
  else  
    can_drv_stat = "";
  }
else if (msg->MsgOV)
  {
  err_nr = msg->MsgData[0];
  lost_msgs = msg->MsgData[1] & ((uint32_t)msg->MsgData[2] << 8);          
  }      
else
  return;
     
if (enable_source)
  {   
  if (msg->MsgErr)
    {     
    if ((err_nr > 0) && (err_nr < 7) && (bus_stat < 4))
      fprintf(LogFile, "      Dev:%2u Ch:%2u  ERROR [%s REC:%u TEC:%u%s] %s\n", dev, ch, CanMsgBusStatusStrings[bus_stat],
	      msg->MsgData[2], msg->MsgData[3], can_drv_stat, CanMsgBusErrorStrings[err_nr]);
    else  
      fprintf(LogFile, "      Dev:%2u Ch:%2u  ERROR Unbek. Fehler\n", dev, ch);
    }
  else
    {
    if ((err_nr > 0) && (err_nr < 4))
      fprintf(LogFile, "      Dev:%2u Ch:%2u  OV [%u] Messages-Lost: %u\n", dev, ch, err_nr, lost_msgs);
    else
      fprintf(LogFile, "      Dev:%2u Ch:%2u  OV Unbek. Fehler\n", dev, ch);
    }    
  }
else
  {
  if (msg->MsgErr)
    {     
    if ((err_nr > 0) && (err_nr < 7) && (bus_stat < 4))
      fprintf(LogFile, "      ERROR [%s REC:%u TEC:%u%s] %s\n",  CanMsgBusStatusStrings[bus_stat], msg->MsgData[2], 
          msg->MsgData[3], can_drv_stat, CanMsgBusErrorStrings[err_nr]);
    else  
      fprintf(LogFile, "      ERROR Unbek. Fehler\n");
    }
  else
    {
    if ((err_nr > 0) && (err_nr < 4))
      fprintf(LogFile, "      OV [%u] Messages-Lost: %u\n", err_nr, lost_msgs);
    else
      fprintf(LogFile, "      OV Unbek. Fehler\n");
    }    
  }  
}  
  

static void LogWriteFdMessagesFull(const char *title, uint32_t log_type, struct TCanFdMsg *msg, int32_t count, int32_t enable_source)
{
uint32_t first;

first = 1;
log_type &= LogFlags;
for (; count; count--)
  {
  if (msg->MsgErr)
    {
    if (log_type & LOG_ERR_MSG)
      {
      if ((title) && (first))
        {
        first = 0;
        LogTimeStamp();
        fprintf(LogFile, "%s\n", title);
        }
      LogWriteFdErrMsg(msg, enable_source);
      }
    }
  else if (msg->MsgOV)
    {      
    if (log_type & LOG_OV_MSG)
      {
      if ((title) && (first))
        {
        first = 0;
        LogTimeStamp();
        fprintf(LogFile, "%s\n", title);
        }      
      LogWriteFdErrMsg(msg, enable_source);
      }
    }
  else
    {
    if (log_type & LOG_TX_MSG)
      {
      if ((title) && (first))
        {
        first = 0;
        LogTimeStamp();
        fprintf(LogFile, "%s\n", title);
        }      
      LogWriteFdMessages(msg, 1, enable_source | 0x02);
      }
    else if (log_type & LOG_RX_MSG)
      {
      if ((title) && (first))
        {
        first = 0;
        LogTimeStamp();
        fprintf(LogFile, "%s\n", title);
        }
      LogWriteFdMessages(msg, 1, enable_source);
      }
    }
  msg++; 
  }  
}


/*
******************** LogCanMessages ********************
*/
void LogCanMessages(uint32_t log_type, uint32_t log_flags, struct TCanMsg *msg, int32_t count)
{
if ((LogEnable) && (log_type & LogFlags))
  {
  if ((!log_flags) || (log_flags & LOG_LOCK))
    DRV_LOCK_ENTER(LogLock);
  LogWriteMessages(msg, count, 0);
#ifndef LOG_FILE_MODE_BUFFERED
  if (!(LogFlags & LOG_DISABLE_SYNC))
    fflush(LogFile);
#endif
  if ((!log_flags) || (log_flags & LOG_UNLOCK))
    DRV_LOCK_LEAVE(LogLock);
  }
}


/*
******************** LogCanMessage ********************
*/
void LogCanMessage(uint32_t log_type, struct TCanMsg *msg, const char *format_str, ...)
{
va_list ap;

if ((LogEnable) && (log_type & LogFlags))
  {
  DRV_LOCK_ENTER(LogLock);
  LogTimeStamp();
  va_start(ap, format_str);
  vfprintf(LogFile, format_str, ap);
  va_end(ap);
  fprintf(LogFile, "\n");
  LogWriteMessages(msg, 1, 0);
#ifndef LOG_FILE_MODE_BUFFERED
  if (!(LogFlags & LOG_DISABLE_SYNC))
    fflush(LogFile);
#endif
  DRV_LOCK_LEAVE(LogLock);
  }
}


/*
******************** LogCanMessagesWithSource ********************
*/
void LogCanMessagesWithSource(uint32_t log_type, uint32_t log_flags, struct TCanMsg *msg, int32_t count)
{
if ((LogEnable) && (log_type & LogFlags))
  {
  if ((!log_flags) || (log_flags & LOG_LOCK))
    DRV_LOCK_ENTER(LogLock);
  LogTimeStamp();  
  fprintf(LogFile, "\n");  
  LogWriteMessages(msg, count, 1);
#ifndef LOG_FILE_MODE_BUFFERED
  if (!(LogFlags & LOG_DISABLE_SYNC))
    fflush(LogFile);
#endif
  if ((!log_flags) || (log_flags & LOG_UNLOCK))
    DRV_LOCK_LEAVE(LogLock);
  }
}


/*
******************** LogCanMessageWithSource ********************
*/
void LogCanMessageWithSource(uint32_t log_type, struct TCanMsg *msg, const char *format_str, ...)
{
va_list ap;

if ((LogEnable) && (log_type & LogFlags))
  {
  DRV_LOCK_ENTER(LogLock);
  LogTimeStamp();
  va_start(ap, format_str);
  vfprintf(LogFile, format_str, ap);
  va_end(ap);
  fprintf(LogFile, "\n"); 
  LogWriteMessages(msg, 1, 1);
#ifndef LOG_FILE_MODE_BUFFERED
  if (!(LogFlags & LOG_DISABLE_SYNC))
    fflush(LogFile);
#endif
  DRV_LOCK_LEAVE(LogLock);
  }
}


/*
******************** LogCanFdMessages ********************
*/
void LogCanFdMessages(uint32_t log_type, uint32_t log_flags, struct TCanFdMsg *msg, int32_t count)
{
if ((LogEnable) && (log_type & LogFlags))
  {
  if ((!log_flags) || (log_flags & LOG_LOCK))
    DRV_LOCK_ENTER(LogLock);  
  LogWriteFdMessagesFull(NULL, log_type, msg, count, 0);
#ifndef LOG_FILE_MODE_BUFFERED
  if (!(LogFlags & LOG_DISABLE_SYNC))
    fflush(LogFile);
#endif
  if ((!log_flags) || (log_flags & LOG_UNLOCK))
    DRV_LOCK_LEAVE(LogLock);
  }
}


/*
******************** LogCanFdMessage ********************
*/
void LogCanFdMessage(uint32_t log_type, struct TCanFdMsg *msg, const char *format_str, ...)
{
va_list ap;
char buf[41];

if ((LogEnable) && (log_type & LogFlags))
  {
  DRV_LOCK_ENTER(LogLock);
  va_start(ap, format_str);
  safe_vsprintf(buf, 40, format_str, ap);
  va_end(ap);
  LogWriteFdMessagesFull(buf, log_type, msg, 1, 0);
#ifndef LOG_FILE_MODE_BUFFERED
  if (!(LogFlags & LOG_DISABLE_SYNC))
    fflush(LogFile);
#endif
  DRV_LOCK_LEAVE(LogLock);
  }
}


/*
******************** LogCanFdMessagesWithSource ********************
*/
void LogCanFdMessagesWithSource(uint32_t log_type, uint32_t log_flags, struct TCanFdMsg *msg, int32_t count)
{
if ((LogEnable) && (log_type & LogFlags))
  {
  if ((!log_flags) || (log_flags & LOG_LOCK))
    DRV_LOCK_ENTER(LogLock);
  LogWriteFdMessagesFull(NULL, log_type, msg, count, 1);
#ifndef LOG_FILE_MODE_BUFFERED
  if (!(LogFlags & LOG_DISABLE_SYNC))
    fflush(LogFile);
#endif
  if ((!log_flags) || (log_flags & LOG_UNLOCK))
    DRV_LOCK_LEAVE(LogLock);
  }
}


/*
******************** LogCanFdMessageWithSource ********************
*/
void LogCanFdMessageWithSource(uint32_t log_type, struct TCanFdMsg *msg, const char *format_str, ...)
{
va_list ap;
char buf[41];

if ((LogEnable) && (log_type & LogFlags))
  {
  DRV_LOCK_ENTER(LogLock);  
  va_start(ap, format_str);
  safe_vsprintf(buf, 40, format_str, ap);
  va_end(ap);  
  LogWriteFdMessagesFull(buf, log_type, msg, 1, 1);
#ifndef LOG_FILE_MODE_BUFFERED
  if (!(LogFlags & LOG_DISABLE_SYNC))
    fflush(LogFile);
#endif
  DRV_LOCK_LEAVE(LogLock);
  }
}


/*
******************** LogStringList ********************
  [Bezeichner]=[Value];[Bezeichner]=[Value];[..]=[..]

*/
void LogStringList(uint32_t log_type, uint32_t log_flags, char *str_list)
{
char *tmp_str, *str, *key, *value;
int32_t res;

if ((LogEnable) && (log_type & LogFlags))
  {
  if ((!log_flags) || (log_flags & LOG_LOCK))
    DRV_LOCK_ENTER(LogLock);
  if (!check_string(str_list))
    fprintf(LogFile, "keine\n");
  else
    {        
    tmp_str = mhs_strdup(str_list);
    str = tmp_str;
    while (*str)
      {
      // **** Key auslesen
      key = get_item_as_string(&str, "=;", &res);
      if ((res < 0) || (!key))
        break;    
      // Fhrende Leerzeichen lschen
      while (*key == ' ')
        key++;
      if (*key == '\0')
        break;    
      // **** Value auslesen
      value = get_item_as_string(&str, "=;", &res);
      if (res == 1) // = gefunden (Key) fehler
        break;
      // Fhrende Leerzeichen lschen
      if (value)
        {
        while (*value == ' ')
          value++;
        if (*value == '\0')
          value = NULL;
        }  
      fprintf(LogFile, "   %s = %s\n", key, value);
      }            
    safe_free(tmp_str);
    }            
#ifndef LOG_FILE_MODE_BUFFERED
  if (!(LogFlags & LOG_DISABLE_SYNC))
    fflush(LogFile);
#endif
  if ((!log_flags) || (log_flags & LOG_UNLOCK))
    DRV_LOCK_LEAVE(LogLock);
  }
}


/*
******************** LogSysError ********************
*/
void LogSysError(char *file_name, char *func_name, int32_t error_no)
{
if ((!file_name) || (!func_name))
  return;
if (LogEnable)
  {
//  if ((!log_flags) || (log_flags & LOG_LOCK))
    DRV_LOCK_ENTER(LogLock);
  LogTimeStamp();  
  fprintf(LogFile, "FATAL-ERROR: File:%s, Funktion:%s, Error:%d\n", file_name, func_name, (int32_t)error_no);
#ifndef LOG_FILE_MODE_BUFFERED
  if (!(LogFlags & LOG_DISABLE_SYNC))
    fflush(LogFile);
#endif
//  if ((!log_flags) || (log_flags & LOG_UNLOCK))
    DRV_LOCK_LEAVE(LogLock);
  }
}


/*
******************** LogHexDump ********************
Funktion  :

Eingaben  :

Ausgaben  : keine
*/
void LogHexDump(uint32_t log_type, uint32_t log_flags, char *data, int32_t count)
{
char str[50];
char *ptr;
int32_t cnt;
unsigned char ch, hex;

if ((LogEnable) && (log_type & LogFlags))
  {
  if ((!log_flags) || (log_flags & LOG_LOCK))
    DRV_LOCK_ENTER(LogLock);
  while (count)
    {
    if (count > 16)
      cnt = 16;
    else
      cnt = count;
    count -= cnt;
    ptr = str;
    for (;cnt;cnt--)
      {
      ch = (unsigned char)*data++;
      hex = ch >> 4;
      if (hex > 9)
        *ptr++ = 'A' + hex;
      else
        *ptr++ = '0' + hex;
      hex = ch & 0x0F;
      if (hex > 9)
        *ptr++ = 'A' + hex;
      else
        *ptr++ = '0' + hex;
      *ptr++ = ' ';
      }
    *ptr = '\0';
    fprintf(LogFile, "   %s\n", str);
    }
#ifndef LOG_FILE_MODE_BUFFERED
  if (!(LogFlags & LOG_DISABLE_SYNC))
    fflush(LogFile);
#endif
  if ((!log_flags) || (log_flags & LOG_UNLOCK))
    DRV_LOCK_LEAVE(LogLock);
  }
}


/*
******************** LogPrintDrvStatus ********************
*/
void LogPrintDrvStatus(uint32_t log_type, uint32_t log_flags, uint32_t channel, struct TDeviceStatus *dev_status)
{
int32_t drv_status;
unsigned char can_status;
unsigned char fifo_status;

drv_status = dev_status->DrvStatus;
can_status = dev_status->CanStatus;
fifo_status = dev_status->FifoStatus;
if ((LogEnable) && (log_type & LogFlags))
  {
  LogTimeStamp();
  if (drv_status > DRV_STATUS_MAX)
    drv_status = DRV_STATUS_MAX;
  if (can_status > CAN_STATUS_MAX)
    can_status = CAN_STATUS_MAX;
  if (fifo_status > FIFO_STATUS_MAX)
    fifo_status = FIFO_STATUS_MAX;
  if ((!log_flags) || (log_flags & LOG_LOCK))
    DRV_LOCK_ENTER(LogLock);
  if (can_status & BUS_FAILURE)    
    fprintf(LogFile, "STATUS: Ch %d: Drv=%s, Can=%s [BUS-FAILURE], Fifo=%s\n", channel, DrvStatusStrings[drv_status],
      CanStatusStrings[can_status & 0x0F], FifoStatusStrings[fifo_status]);
  else      
    fprintf(LogFile, "STATUS: Ch %d: Drv=%s, Can=%s, Fifo=%s\n", channel, DrvStatusStrings[drv_status],
      CanStatusStrings[can_status & 0x0F], FifoStatusStrings[fifo_status]);
#ifndef LOG_FILE_MODE_BUFFERED
  if (!(LogFlags & LOG_DISABLE_SYNC))
    fflush(LogFile);
#endif
  if ((!log_flags) || (log_flags & LOG_UNLOCK))
    DRV_LOCK_LEAVE(LogLock);
  }
}


void LogPrintUsbDevices(uint32_t log_type, uint32_t log_flags, struct TCanDevicesList *list, int32_t num_devs)
{
int32_t i;

if (!list)
  num_devs = 0;
if ((LogEnable) && (log_type & LogFlags))
  {
  if ((!log_flags) || (log_flags & LOG_LOCK))
    DRV_LOCK_ENTER(LogLock);
  fprintf(LogFile, "    Open    |Tiny-CAN Id | Snr        | USB-Description \n");
  fprintf(LogFile, "============+============+============+=====================\n");
  if (num_devs <= 0)
    fprintf(LogFile, "keine Devices gefunden.\n");
  else
    {
    for (i = 0; i < num_devs; i++)
      {
      fprintf(LogFile, " 0x%08X | 0x%08X | %10s | %s\n", list[i].TCanIdx, list[i].HwId,
          list[i].SerialNumber, list[i].Description);
      }
    }
  fprintf(LogFile, "============+============+============+=====================\n");
#ifndef LOG_FILE_MODE_BUFFERED
  if (!(LogFlags & LOG_DISABLE_SYNC))
    fflush(LogFile);
#endif
  if ((!log_flags) || (log_flags & LOG_UNLOCK))
    DRV_LOCK_LEAVE(LogLock);
  }
}


static const char *GetValueTypeString(uint32_t type)
{
int32_t i;
const struct TValueTypeDesc *item;

for (i = 0; i < VALUE_TYPE_DESC_SIZE; i++)
  {
  item = &ValueTypeDesc[i];
  if (item->Type == type)
    return(item->Bezeichner);
  }
return(NULL);
}


void LogPrintHwInfo(uint32_t log_type, uint32_t log_flags, struct TInfoVar *info, uint32_t size)
{
uint32_t i, value, tag, monat, jahr, major,  minor, revision;

if ((LogEnable) && (log_type & LogFlags))
  {
  if ((!log_flags) || (log_flags & LOG_LOCK))
    DRV_LOCK_ENTER(LogLock);
    
  fprintf(LogFile, " Key   |   Type   | Size | Value\n");
  fprintf(LogFile, "-------+----------+------+-------------------------------\n");
  for (i = 0; i < size; i++)
    {  
    fprintf(LogFile, "0x%04X | %-8s |  %-3u | ", info[i].Key, GetValueTypeString(info[i].Type), info[i].Size);

    switch (info[i].Type)
      {
      case VT_BYTE   : {
                       fprintf(LogFile, "%d\n", *((signed char*)(info[i].Data)));
                       break;
                       }
      case VT_UBYTE  : {
                       fprintf(LogFile, "%u\n", *((unsigned char *)(info[i].Data)));
                       break;
                       }
      case VT_HBYTE  : {
                       fprintf(LogFile, "0x%02X\n", *((unsigned char *)(info[i].Data)));
                       break;
                       }
      case VT_WORD   : {
                       fprintf(LogFile, "%d\n", *((signed short *)(info[i].Data)));
                       break;
                       }
      case VT_UWORD  : {
                       fprintf(LogFile, "%u\n", *((uint16_t *)(info[i].Data)));
                       break;
                       }
      case VT_HWORD  : {
                       fprintf(LogFile, "0x%04X\n", *((uint16_t *)(info[i].Data)));
                       break;
                       }
      case VT_LONG   : {
                       fprintf(LogFile, "%d\n", *((int32_t *)(info[i].Data)));
                       break;
                       }
      case VT_ULONG  : {
                       fprintf(LogFile, "%u\n", *((uint32_t *)(info[i].Data)));
                       break;
                       }
      case VT_HLONG  : {
                       fprintf(LogFile, "0x%08X\n", *((uint32_t *)(info[i].Data)));
                       break;
                       }                
      case VT_REVISION :
                       {
                       value = *((uint32_t *)(info[i].Data));
                       major = value / 10000;
                       value = value % 10000;  
                       minor = value / 100;                   
                       revision = value % 100;                   
                       fprintf(LogFile, "%u.%u.%u\n", major, minor, revision); 
                       break; 
                       }
      case VT_DATE   : {
                       value = *((uint32_t *)(info[i].Data));
                       tag = GetBCD(value & 0xFF, 2);
                       monat = GetBCD((value >> 8) & 0xFF, 2);  
                       jahr = GetBCD((value >> 16) & 0xFFFF, 4);
                       fprintf(LogFile, "%02u.%02u.%4u\n", tag, monat, jahr);
                       break; 
                       }                       
      case VT_STRING : {                    
                       fprintf(LogFile, "%s\n", info[i].Data);
                       break;
                       }
      }
    }
  fprintf(LogFile, "-------+----------+------+-------------------------------\n");  
#ifndef LOG_FILE_MODE_BUFFERED
  if (!(LogFlags & LOG_DISABLE_SYNC))
    fflush(LogFile);
#endif
  if ((!log_flags) || (log_flags & LOG_UNLOCK))
    DRV_LOCK_LEAVE(LogLock);  
  }    
}


void LogPrintDeviceInfo(uint32_t log_type, uint32_t log_flags, struct TCanDeviceInfo *info)
{
char *str;
char str_puf[100];
uint32_t ver, ver2;

if ((LogEnable) && (log_type & LogFlags))
  {
  if ((!log_flags) || (log_flags & LOG_LOCK))
    DRV_LOCK_ENTER(LogLock);
  fprintf(LogFile, "Device:%s [Snr:%s]\n", info->Description, info->SerialNumber);
  fprintf(LogFile, "   Hardware Id      : 0x%08X\n", info->HwId);
  ver = info->FirmwareVersion / 1000;
  ver2 = info->FirmwareVersion % 1000;
  fprintf(LogFile, "   Firmware Version : %u.%03u\n", ver, ver2);
  fprintf(LogFile, "   CanClock         : %u\n", info->ModulFeatures.CanClock);
  str = str_puf;
  if (info->ModulFeatures.Flags & CAN_FEATURE_LOM)            // Silent Mode (LOM = Listen only Mode)
    str = mhs_stpcpy(str, "LOM ");
  if (info->ModulFeatures.Flags & CAN_FEATURE_ARD)            // Automatic Retransmission disable
    str = mhs_stpcpy(str, "ARD ");
  if (info->ModulFeatures.Flags & CAN_FEATURE_TX_ACK)         // TX ACK
    str = mhs_stpcpy(str, "TX_ACK ");
  if (info->ModulFeatures.Flags & CAN_FEATURE_HW_TIMESTAMP)
    str = mhs_stpcpy(str, "HW_TIMESTAMP ");
  fprintf(LogFile, "   Features-Flags   : %s\n", str_puf);
  fprintf(LogFile, "   CanChannelsCount : %u\n", info->ModulFeatures.CanChannelsCount);
  fprintf(LogFile, "   HwRxFilterCount  : %u\n", info->ModulFeatures.HwRxFilterCount);
  fprintf(LogFile, "   HwTxPufferCount  : %u\n", info->ModulFeatures.HwTxPufferCount);
  fprintf(LogFile, "\n");
#ifndef LOG_FILE_MODE_BUFFERED
  if (!(LogFlags & LOG_DISABLE_SYNC))
    fflush(LogFile);
#endif
   if ((!log_flags) || (log_flags & LOG_UNLOCK))
    DRV_LOCK_LEAVE(LogLock);  
  }
}
