/***************************************************************************
                         tcan_setup.c  -  description
                             -------------------
    begin             : 10.05.2017
    last modify       : 30.05.2020
    copyright         : (C) 2017 - 2020 by MHS-Elektronik GmbH & Co. KG, Germany
    author            : Klaus Demlehner, klaus@mhs-elektronik.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software, you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License           *
 *   version 2.1 as published by the Free Software Foundation.             *
 *                                                                         *
 ***************************************************************************/

/**
    Library to talk to Tiny-CAN devices. You find the latest versions at
       http://www.tiny-can.com/
**/
#include "global.h"
#include <stdio.h>
#include <string.h>
#include "util.h"
#include "errors.h"
#include "log.h"
#include "can_core.h"
#include "can_puffer.h"
#include "mhs_obj.h"
#include "var_obj.h"
#include "tar_drv.h"
#include "tar_drv2.h"
#include "fd_speeds.h"
#include "tcan_setup.h"



#define CAN_VAR_SPEED_FLAG             0x01
#define CAN_VAR_SPEED_BTR_FLAG         0x02
#define CAN_VAR_TX_ACK_ENABLE_FLAG     0x04
#define CAN_VAR_ERROR_LOG_ENABLE_FLAG  0x08
#define CAN_VAR_TIMESTAMP_ENABLE_FLAG  0x10
#define CAN_VAR_SPEED_DBTR_FLAG        0x20                                      
#define CAN_VAR_FD_FLAGS_FLAG          0x40
#define CAN_VAR_FIFO_OV_MODE_FLAG      0x80  


/*

mode -> Flags
         SETUP_LOCAL   = Keine Daten zur Hardware bertragen
         SETUP_SET_ALL = Alle Setup Variablen setzen
         SETUP_CAN_STOP = CAN-Bus stoppen
         SETUP_CAN_CLEANUP = Rx & Tx FIFO lschen, Rx-OV & Tx-OV lschen

*/
int32_t ProcessSetup(TCanDevice *dev, uint32_t mode)
{
int32_t err;
TMhsObj *obj;
TObjValue *value;
TMhsObjContext *context, *main_context;
#if RX_FILTER_MAX > 0
TMhsObjContext *filter_context;
#endif
#if TX_PUFFER_MAX > 0
TMhsObjContext *tx_puffer_context;
struct TCanMsg msg;
#endif
uint8_t channels, channel, idx, set_var_flags_global, time_stamp_mode,
   global_tx_ack_enable, global_can_err_msgs_enable, set_io_cfg;
uint16_t mask, set_puffer, speed;
uint8_t set_var_flags[MAX_CAN_CHANNELS];
uint16_t can_speed[MAX_CAN_CHANNELS];
uint32_t can_speed_user[MAX_CAN_CHANNELS];
uint8_t tx_ack_enable[MAX_CAN_CHANNELS];
uint8_t can_err_msgs_enable[MAX_CAN_CHANNELS];
uint16_t set_filter[MAX_CAN_CHANNELS];
uint32_t io_cfg[MAX_IO_CFG];
uint32_t can_dbtr[MAX_CAN_CHANNELS];    
uint8_t can_fd_flags[MAX_CAN_CHANNELS];
union TCanIndex index;
uint32_t cmd, i;
char str[26];
#if (TX_PUFFER_MAX > 0) || (RX_FILTER_MAX > 0)
TObjCanPuffer *puffer;
#endif

main_context = can_main_get_context();
context = dev->Context;
if (mode & (SETUP_SET_ALL | SETUP_LOCAL))
  {
  // **** Alle Setup Variablen setzen
  dev->OvMode = mhs_value_get_as_uword("FifoOvMode", ACCESS_INIT, main_context);  
  if (mhs_value_get_status("FilterReadIntervall", context) & MHS_VAL_FLAG_WRITE)
    dev->FilterReadIntervall = mhs_value_get_as_ubyte("FilterReadIntervall", ACCESS_INIT, context);
  else
    dev->FilterReadIntervall = mhs_value_get_as_ubyte("FilterReadIntervall", ACCESS_INIT, main_context);
  if (mhs_value_get_status("HighPollIntervall", context) & MHS_VAL_FLAG_WRITE)
    dev->HighPollIntervall = mhs_value_get_as_ubyte("HighPollIntervall", ACCESS_INIT, context);
  else
    dev->HighPollIntervall = mhs_value_get_as_ubyte("HighPollIntervall", ACCESS_INIT, main_context);
  if (mhs_value_get_status("IdlePollIntervall", context) & MHS_VAL_FLAG_WRITE)
    dev->IdlePollIntervall = mhs_value_get_as_ubyte("IdlePollIntervall", ACCESS_INIT, context);
  else
    dev->IdlePollIntervall = mhs_value_get_as_ubyte("IdlePollIntervall", ACCESS_INIT, main_context);    
  
  if (mhs_value_get_status("TxCanFifoEventLimit", context) & MHS_VAL_FLAG_WRITE)
    dev->TxCanFifoEventLimit = mhs_value_get_as_ubyte("TxCanFifoEventLimit", ACCESS_INIT, context);
  else
    dev->TxCanFifoEventLimit = mhs_value_get_as_ubyte("TxCanFifoEventLimit", ACCESS_INIT, main_context);
  if (mhs_value_get_status("CommTryCount", context) & MHS_VAL_FLAG_WRITE)
    dev->CommTryCount = mhs_value_get_as_ubyte("CommTryCount", ACCESS_INIT, context);
  else
    dev->CommTryCount = mhs_value_get_as_ubyte("CommTryCount", ACCESS_INIT, main_context);  
  }

if (mode & SETUP_LOCAL)
  return(0);
if ((!dev->ModulDesc) || (dev->DeviceStatus.DrvStatus < DRV_STATUS_PORT_OPEN))
  return(0);
#if RX_FILTER_MAX > 0
filter_context = dev->RxFilterContext;
#endif
#if TX_PUFFER_MAX > 0
tx_puffer_context = dev->TxPufferContext;
#endif
set_io_cfg = 0x00;
channels = (uint8_t)dev->ModulDesc->CanChannels;
if (mode & SETUP_SET_ALL)
  {
  for (channel = 0; channel < channels; channel++)
    {
    set_var_flags[channel] = 0xFF;
    set_filter[channel] = 0xFFFF;
    }
  set_puffer = 0xFFFF;
  set_var_flags_global = 0xFF;
  }
else
  {
  for (channel = 0; channel < channels; channel++)
    {
    set_var_flags[channel] = 0x00;
    set_filter[channel] = 0x0000;
    }
  set_puffer = 0x0000;
  set_var_flags_global = 0x00;
  }
for (i = 50; i; i--)
  {
  if (!(obj = mhs_object_get_cmd_marked(context)))
    break;
  channel = (uint8_t)obj->UserUInt & 0xFF;
  idx = (uint8_t)(obj->UserUInt >> 8);
  cmd = obj->UserUInt & 0xFFFF0000;
  value = (TObjValue *)obj;
  if (cmd == CAN_VAR_OBJ)
    {
    switch (idx)
      {
      case CAN_VAR_SPEED :            {
                                      can_speed[channel] = value->Value.U16;
                                      set_var_flags[channel] |= CAN_VAR_SPEED_FLAG;
                                      break;
                                      }
      case CAN_VAR_SPEED_BTR :        {
                                      safe_sprintf(str, 25, "CanSpeed%u", channel + 1);
                                      (void)mhs_value_set_as_uword(str, 0, 0xFFFF, 0, dev->Context);
                                      can_speed[channel] = 0;
                                      can_speed_user[channel] = value->Value.U32;
                                      set_var_flags[channel] |= CAN_VAR_SPEED_BTR_FLAG;
                                      break;
                                      }
      case CAN_VAR_TX_ACK_ENABLE :    {
                                      tx_ack_enable[channel] = value->Value.U8;
                                      set_var_flags[channel] |= CAN_VAR_TX_ACK_ENABLE_FLAG;
                                      break;
                                      }
      case CAN_VAR_ERROR_LOG_ENABLE : {
                                      can_err_msgs_enable[channel] = value->Value.U8;
                                      set_var_flags[channel] |= CAN_VAR_ERROR_LOG_ENABLE_FLAG;
                                      break;
                                      }                                
      case CAN_VAR_SPEED_DBTR :       {
                                      safe_sprintf(str, 25, "CanDSpeed%u", channel + 1);
                                      (void)mhs_value_set_as_uword(str, 0, 0xFFFF, 0, dev->Context);                                      
                                      can_dbtr[channel] = value->Value.U32;
                                      set_var_flags[channel] |= CAN_VAR_SPEED_DBTR_FLAG;
                                      break;                                                               
                                      }
      case CAN_VAR_FD_FLAGS :         {                             
                                      can_fd_flags[channel] = value->Value.U8;
                                      set_var_flags[channel] |= CAN_VAR_FD_FLAGS_FLAG;
                                      }                                                                         
      }
    }
  else if (cmd == IVAR_OBJ)
    {
    switch (idx)
      {
      case IVAR_D_SPEED :             {
                                      can_dbtr[channel] = GetDBtrValueFromSpeed(dev, value->Value.U16);
                                      set_var_flags[channel] |= CAN_VAR_SPEED_DBTR_FLAG;
                                      break;
                                      }
      }
    }  
  else
    {
    switch (cmd)
      {
      case IO_VAR_PORT_OUT        : {
                                    dev->DigOut[channel] = value->Value.U16;
                                    dev->DigOutSet |= (0x01 << channel);
                                    break;
                                    }
      case IO_VAR_PORT_ANALOG_OUT : {
                                    dev->AnalogOut[channel] = value->Value.U16;
                                    dev->AnalogOutSet |= (0x01 << channel);
                                    break;
                                    }
      case IO_VAR_CFG             : {
                                    io_cfg[channel] = value->Value.U32;
                                    set_io_cfg |= (0x01 << channel);
                                    break;
                                    }
      case FILTER_READ_INTERVALL  : {
                                    dev->FilterReadIntervall = value->Value.U32;
                                    break;
                                    }
      case HIGH_POLL_INTERVALL    : {
                                    dev->HighPollIntervall = value->Value.U32;
                                    break;
                                    }
      case IDLE_POLL_INTERVALL    : {
                                    dev->IdlePollIntervall = value->Value.U32;
                                    break;
                                    }
      case TX_CAN_FIFO_EVENT_LIMIT: {
                                    dev->TxCanFifoEventLimit = value->Value.U32;
                                    break;
                                    }
      case COMM_TRY_COUNT         : {
                                    dev->CommTryCount = value->Value.U8;
                                    break;
                                    }                               
      }
    }
  }
// **** Filter
#if RX_FILTER_MAX > 0
for (i = 32; i; i--)
  {
  if (!(obj = mhs_object_get_cmd_marked(filter_context)))
    break;
  index.Long = obj->Index;
  if (index.Long & (INDEX_FIFO_VIRTUAL | INDEX_SOFT_FLAG | INDEX_TXT_FLAG))
    continue;
  channel = index.Item.Channel;
  set_filter[channel] |= (0x0001 << (index.Item.SubIndex - 1));
  }
#endif
#if TX_PUFFER_MAX > 0
if (dev->Protokoll != 2)
  {
  for (i = 32; i; i--)
    {
    if (!(obj = mhs_object_get_cmd_marked(tx_puffer_context)))
      break;
    index.Long = obj->Index;
    set_puffer |= (0x0001 << (index.Item.SubIndex - 1));
    }
  }
#endif

if (mode & SETUP_SET_ALL)
  {  
  global_tx_ack_enable = mhs_value_get_as_ubyte("CanTxAckEnable", ACCESS_INIT, context);
  global_can_err_msgs_enable = mhs_value_get_as_ubyte("CanErrorMsgsEnable", ACCESS_INIT, context);
  for (i = 0; i < channels; i++)
    {
    safe_sprintf(str, 25, "CanSpeed%u", i + 1);
    can_speed[i] = mhs_value_get_as_uword(str, ACCESS_INIT, context);
    safe_sprintf(str, 25, "CanSpeed%uUser", i + 1);
    can_speed_user[i] = mhs_value_get_as_ulong(str, ACCESS_INIT, context);
    safe_sprintf(str, 25, "CanDSpeed%u", i + 1);
    speed = mhs_value_get_as_uword(str, ACCESS_INIT, context);
    if (speed)
      can_dbtr[i] = GetDBtrValueFromSpeed(dev, speed);
    else
      {  
      safe_sprintf(str, 25, "CanDSpeed%uUser", i + 1);
      can_dbtr[i] = mhs_value_get_as_ulong(str, ACCESS_INIT, context);
      }
    safe_sprintf(str, 25, "CanFdFlags%u", i + 1);
    can_fd_flags[i] = mhs_value_get_as_ubyte(str, ACCESS_INIT, context);
        
    safe_sprintf(str, 25, "Can%uTxAckEnable", i + 1);
    if (mhs_value_get_status(str, context) & MHS_VAL_FLAG_WRITE)
      tx_ack_enable[i] = mhs_value_get_as_ubyte(str, ACCESS_INIT, context);
    else
      tx_ack_enable[i] = global_tx_ack_enable;
    safe_sprintf(str, 25, "Can%uErrorMsgsEnable", i + 1);
    if (mhs_value_get_status(str, context) & MHS_VAL_FLAG_WRITE)
      can_err_msgs_enable[i] = mhs_value_get_as_ubyte(str, ACCESS_INIT, context);
    else
      can_err_msgs_enable[i] = global_can_err_msgs_enable;
    }
  if (mhs_value_get_status("TimeStampMode", context) & MHS_VAL_FLAG_WRITE)
    time_stamp_mode = mhs_value_get_as_ubyte("TimeStampMode", ACCESS_INIT, context);
  else
    time_stamp_mode = mhs_value_get_as_ubyte("TimeStampMode", ACCESS_INIT, main_context);
  if (dev->Protokoll == 2)
    {
    if ((i = dev->ModulDesc->DigIoCount))
      {
      idx = (uint8_t)(i / 8);
      if ((i % 8))
        idx++;
      set_io_cfg = 0xFF;
      for (channel = 0; channel < idx; channel++)
        {
        safe_sprintf(str, 25, "IoCfg%u", channel);
        io_cfg[channel] = mhs_value_get_as_ulong(str, ACCESS_PUBLIC, context);
        }

      idx = (uint8_t)(i / 16);
      if ((i % 16))
        idx++;
      dev->DigOutSet = 0;
      for (channel = 0; channel < idx; channel++)
        {
        safe_sprintf(str, 25, "IoPort%uOut", channel);
        dev->DigOut[channel] = mhs_value_get_as_uword(str, ACCESS_PUBLIC, context);
        dev->DigOutSet |= (0x01 << channel);
        }      
      for (channel = 0; channel < 4; channel++)
        {
        safe_sprintf(str, 25, "IoPort%uAnalogOut", channel);
        dev->AnalogOut[channel] = mhs_value_get_as_uword(str, ACCESS_PUBLIC, context);
        dev->AnalogOutSet |= (0x01 << channel);
        }
      }
    }
  }

if ((dev->Protokoll == 2) && ((i = dev->ModulDesc->DigIoCount)))
  {
  if (set_io_cfg)
    {
    idx = (uint8_t)(i / 8);
    if ((i % 8))
      idx++;
    dev->UseDigIo = 0;
    for (channel = 0; channel < idx; channel++)
      {
      safe_sprintf(str, 25, "IoCfg%u", channel);
      if (mhs_value_get_as_ulong(str, ACCESS_PUBLIC, context))
        {
        dev->UseDigIo = 1;
        break;
        }
      }
    }
  }
else
  dev->UseDigIo = 0;

if (mode & SETUP_CAN_STOP)
  {
  if (dev->Protokoll == 2)
    {
    for (channel = 0; channel < channels; channel++)
      {
      if ((err = TAR2SetCANMode(dev, channel, OP_CAN_STOP)))  // CAN-Bus stoppen
        return(err);
      }
    }
  else
    {
    if ((err = TARSetCANMode(dev, OP_CAN_STOP)))  // CAN-Bus stoppen
      return(err);
    }
  }
/*************************************/
/*             Protokoll 2           */
/*************************************/
if (dev->Protokoll == 2)
  {
  // **** Einstellungen bertragen
  for (channel = 0; channel < channels; channel++)
    {
    if (set_var_flags[channel] & (CAN_VAR_SPEED_FLAG | CAN_VAR_SPEED_BTR_FLAG))
      {
      if (can_speed[channel])
        {
        if ((err = TAR2CanSetSpeed(dev, channel, can_speed[channel])))
          return(err);
        }
      else
        {
        if ((err = TAR2CanSetBtr(dev, channel, can_speed_user[channel])))
          return(err);
        }
      }
    }
  if (dev->ModulDesc->CanFeaturesFlags & CAN_FEATURE_FD_HARDWARE)
    { 
    for (channel = 0; channel < channels; channel++)
      {            
      if (dev->CanFd)
        {          
        if (set_var_flags[channel] & CAN_VAR_SPEED_DBTR_FLAG)                                      
          {
          if ((err = TAR2CanSetDBtr(dev, channel, can_dbtr[channel])))
            return(err);
          }  
        if (set_var_flags[channel] & CAN_VAR_FD_FLAGS_FLAG)
          {    
          if ((err = TAR2CanSetFdFlags(dev, channel, can_fd_flags[channel])))
            return(err);
          }
        }
      else
        {
        if (set_var_flags[channel] & CAN_VAR_FD_FLAGS_FLAG)
          {    
          if ((err = TAR2CanSetFdFlags(dev, channel, 0)))
            return(err);
          }
        }      
      }
    }    
  if (dev->ModulDesc->SupportTxAck)
    {
    for (channel = 0; channel < channels; channel++)
      {
      if (set_var_flags[channel] & CAN_VAR_TX_ACK_ENABLE_FLAG)
        {
        if ((err = TAR2CanSetTxAckEnable(dev, channel, tx_ack_enable[channel])))
          return(err);
        }
      }
    }
  if (dev->ModulDesc->SupportCanErrorMsgs)
    {
    for (channel = 0; channel < channels; channel++)
      {
      if (set_var_flags[channel] & CAN_VAR_ERROR_LOG_ENABLE_FLAG)
        {
        if ((err = TAR2CanSetErrorLogEnable(dev, channel, can_err_msgs_enable[channel])))
          return(err);
        }
      }
    }
  if (set_var_flags_global & CAN_VAR_TIMESTAMP_ENABLE_FLAG)
    {
    dev->TimeStampMode = time_stamp_mode;
    if (dev->ModulDesc->SupportHwTimestamp)
      {
      // **** Hardware Time Stamps einstellen      
      if (time_stamp_mode >= TIME_STAMP_HW_UNIX)
        i = 1;
      else
        i = 0;
      if ((err = TAR2SetTimeStamp(dev, (unsigned char)i)))
        return(err);
      }
    }   
  if (dev->ModulDesc->CanFeaturesFlags & CAN_FEATURE_FIFO_OV_MODE)
    {  
    if (set_var_flags_global & CAN_VAR_FIFO_OV_MODE_FLAG)
      {   
      if ((err = TAR2SetFifoOvMode(dev, (uint8_t)(dev->OvMode & 0xFF))))
        return(err);
      }
    }     
#if RX_FILTER_MAX > 0
  // **** Filter einstellen
  for (channel = 0; channel < channels; channel++)
    {
    mask = 0x0001;
    index.Long =  dev->Index;
    for (i = 1; i <= dev->ModulDesc->HwRxFilterCount; i++)
      {
      if (mask & set_filter[channel])
        {
        index.Item.Channel = channel;
        index.Item.SubIndex = i;
        if ((puffer = (TObjCanPuffer *)mhs_object_get_by_index(index.Long, filter_context)))
          {
          if ((err = TAR2SetFilter(dev, channel, i, &((TObjCan *)puffer)->Filter)) < 0)
            return(err);
          }
        }
      mask <<= 1;
      }
    }
#endif
  if ((i = dev->ModulDesc->DigIoCount))
    {
    idx = (uint8_t)(i / 8);
    if ((i % 8))
      idx++;
    mask = 0x0001;
    for (channel = 0; channel < idx; channel++)
      {
      if (mask & set_io_cfg)
        {
        if ((err = TAR2SetIoConfig(dev, channel, io_cfg[channel])) < 0)
          return(err);
        }
      mask <<= 1;
      }
    }
  }
/*************************************/
/*             Protokoll 1           */
/*************************************/
else
  {
  // **** Einstellungen übertragen
  if (set_var_flags[0] & (CAN_VAR_SPEED_FLAG | CAN_VAR_SPEED_BTR_FLAG))
    {
    if (can_speed[0])
      {
      if ((err = TARCanSetSpeed(dev, can_speed[0])))
        return(err);
      }
    else
      {
      if ((err = TARCanSetBtr(dev, can_speed_user[0])))
        return(err);
      }
    }
  if (set_var_flags[0] & CAN_VAR_TX_ACK_ENABLE_FLAG)
    {
    if (dev->ModulDesc->SupportTxAck)
      {
      if ((err = TARCanSetTxAckEnable(dev, tx_ack_enable[0])))
        return(err);
      }
    }
  if (set_var_flags[0] & CAN_VAR_ERROR_LOG_ENABLE_FLAG)
    {
    if (dev->ModulDesc->SupportCanErrorMsgs)
      {
      if ((err = TARCanSetErrorLogEnable(dev, can_err_msgs_enable[0])))
        return(err);
      }
    }
  if (set_var_flags_global & CAN_VAR_TIMESTAMP_ENABLE_FLAG)
    {
    dev->TimeStampMode = time_stamp_mode;
    if (dev->ModulDesc->SupportHwTimestamp)
      {
      // **** Hardware Time Stamps einstellen      
      if (time_stamp_mode >= TIME_STAMP_HW_UNIX)
        i = 1;
      else
        i = 0;
      if ((err = TARSetTimeStamp(dev, (unsigned char)i)))
        return(err);
      }
    }
#if RX_FILTER_MAX > 0
  // **** Filter einstellen
  mask = 0x0001;
  for (i = 1; i <= dev->ModulDesc->HwRxFilterCount; i++)
    {
    if (mask & set_filter[0])
      {
      if ((puffer = (TObjCanPuffer *)mhs_object_get_by_index(i, filter_context)))
        {
        if ((err = TARSetFilter(dev, i, &((TObjCan *)puffer)->Filter)) < 0)
          return(err);
        }
      }
    mask <<= 1;
    }
#endif
#if TX_PUFFER_MAX > 0
  // **** Intervall Boxen setzen
  mask = 0x0001;
  for (i = 1; i <= dev->ModulDesc->HwTxPufferCount; i++)
    {
    if (mask & set_puffer)
      {
      if ((puffer = (TObjCanPuffer *)mhs_object_get_by_index(i | dev->Index | INDEX_TXT_FLAG, tx_puffer_context)))
        {
        if (mhs_can_puffer_get_copy(puffer, &msg) > 0)
          {
          if ((err = TARTxFifoPut(dev, i, &msg, 1)) < 0)
            return(err);
          }
        else
          mhs_can_puffer_set_intervall(puffer, 0);
        if ((err = TARCanSetIntervall(dev, i, mhs_can_puffer_get_intervall(puffer))) < 0)
          return(err);
        }
       else
        {
        if ((err = TARCanSetIntervall(dev, i, 0)) < 0)
          return(err);
        }
      }
    mask <<= 1;
    }
#endif
  }

if (mode & SETUP_CAN_CLEANUP)
  {
  if (dev->Protokoll == 2)
    {
    for (channel = 0; channel < channels; channel++)
      {
      // CAN Stop, Rx & Tx FIFO lschen, Rx-OV & Tx-OV lschen
      if ((err = TAR2SetCANMode(dev, channel, OP_CAN_STOP | 0xA0)))            
        return(err);
      }
    }
  else
    {
    // CAN Stop, Rx & Tx FIFO lschen, Rx-OV & Tx-OV lschen
    if ((err = TARSetCANMode(dev, OP_CAN_STOP | 0xA0))) 
      return(err);
    }
  }
return(0);
}












